﻿<#   ARP Scan v1.1

	 Created by Robert Sappert-Ernst | www.sappert.com

example:
ArpScan.ps1 -fromIp "192.168.0.1" -toIp "192.168.0.254"
ArpScan.ps1 -fromIp "192.168.0.1" -toIp "192.168.0.254" -threadsCount 500

Attention:
You must call a powershell script always with a path like:
	.\ArpScan.ps1 ....
	c:\MyTools\ArpScan.ps1 ...

*****************************************************************************************
Copyright 2024 Robert Sappert-Ernst Consulting
Alle Rechte vorbehalten.
 
Diese Software wird ohne Garantie zur Verfügung gestellt und darf frei verwendet werden, 
auch in kommerziellen Umgebungen, solange dieser Urheberrechtshinweis nicht entfernt wird.

***************************************************************************************** 
Copyright 2024 Robert Sappert-Ernst Consulting

All rights reserved. 

This software is provided without warranty and may be freely used, even in commercial environments, 
as long as this copyright notice is not removed.

*****************************************************************************************#>

param (
    [Parameter(Mandatory=$true)]
    [string]$fromIp,
    [Parameter(Mandatory=$false)]
    [string]$toIp = $fromIp,
    [Parameter(Mandatory=$false)]
    [int]$threadsCount = 100
)

# Set strict mode
Set-StrictMode -Version latest

# Initialize BlockingCollections
$jobQueue = [System.Collections.Concurrent.BlockingCollection[string]]::new()
$resultsQueue = [System.Collections.Concurrent.BlockingCollection[string]]::new()

# Default settings
[int]$maxThreads = $threadsCount;
[int]$maxJobsOnJobQueue = 200;

#ip addresses
[string]$ipStart = $fromIp
[string]$ipEnd = $toIp
[UInt32]$ipStartInt = [UInt32]::MaxValue
[UInt32]$ipEndInt = [UInt32]::MaxValue
[UInt32]$global:currentValue = 0

class MacVendorPair {
    [string]$MacPrefix
    [string]$VendorName
}

$macVendorList = New-Object System.Collections.ArrayList
$basePath = $PSScriptRoot
write-host "Start loading mac-vendors-export.csv"
# Import the CSV and fill the list

try
{
    Import-Csv -Path "$basePath\mac-vendors-export.csv" | ForEach-Object {
	    $pair = [MacVendorPair]::new()
	    $pair.MacPrefix = $_."Mac Prefix".Replace(":", "").ToUpper() # Normalize MAC prefix
	    $pair.VendorName = $_."Vendor Name"
	    $macVendorList.Add($pair) > $null # Adding to ArrayList and suppressing output
    }
} catch {
    $errorMessage = $_.Exception.Message
    write-host $errorMessage
}

write-host "loading mac-vendors-export.csv finished"


if (-not ([System.Management.Automation.PSTypeName]'NativeMethods').Type) {
    Add-Type -TypeDefinition @"
    using System;
    using System.Runtime.InteropServices;

    public static class NativeMethods {
        private const string IphlpApi = "iphlpapi.dll";

        [DllImport(IphlpApi, ExactSpelling = true)]
        public static extern int SendARP(int DestIP, int SrcIP, byte[] pMacAddr, ref int PhyAddrLen);
    }
"@ -ErrorAction SilentlyContinue
}


# Job-Source
# ------------------------------------------------------------
function Get-NextJob {
    if ($global:currentValue -gt $ipEndInt) {
        return $null
    } else {
        return ($global:currentValue++)
    }
}

# from: https://blog.idera.com/database-tools/calculate-broadcast-address
# ------------------------------------------------------------
filter Convert-Decimal2IP {
	([System.Net.IPAddress]$_).IPAddressToString 
}

# Convert an IP to an integer value for compare and iterating
#
# some parts from: https://stackoverflow.com/a/60579253/1574221
# ------------------------------------------------------------
function Ip2Int() {
    param (
        [Parameter(Mandatory=$true)]
        [string]$ipAddress
    )

    # Get the network Address
    $ipDummy = [ipaddress]$ipAddress

    $bytes = $ipDummy.GetAddressBytes()
    if ([BitConverter]::IsLittleEndian) {
	    [Array]::Reverse($bytes)
    }
    $netIpInt = [BitConverter]::ToUInt32($bytes, 0)

    return $netIpInt
}


# Define the script block to be executed in each runspace
# ------------------------------------------------------------
$scriptBlock = {
    param($jobQueue, $resultsQueue, $macVendorList)

    # from: https://blog.idera.com/database-tools/calculate-broadcast-address
    # ------------------------------------------------------------
    filter Convert-Decimal2IP {
	    ([System.Net.IPAddress]$_).IPAddressToString 
    }

    # send the arp package and wait for a response
    function Perform-ARP {
	    param($destIp)
	    $addr = New-Object byte[] 6
	    $len = $addr.Length
	    $res = [NativeMethods]::SendARP($destIp, 0, $addr, [ref]$len)
	    if ($res -eq 0) {
		    $mac = ($addr | ForEach-Object { $_.ToString("X2") }) -join "-"
		    return $mac
	    }
	    return $null
    }

    try {
        while ($true) {
            # Take a job from the queue, blocking if the queue is empty
            $ipAsRegularInteger = $jobQueue.Take()
            #convert back to string ip
            $ipTemp = $ipAsRegularInteger | Convert-Decimal2IP
            #convert to int
            $destIp = [System.BitConverter]::ToInt32([System.Net.IPAddress]::Parse($ipTemp).GetAddressBytes(), 0)
            #get arp-result
            $arpResult = Perform-ARP $destIp
            # if could resolve
            if ($arpResult -ne $null) {

                $normalizedMac = $arpResult.Replace(":", "").ToUpper()
                $normalizedMac = $arpResult.Replace("-", "").ToUpper()
                $vendorName = "<no vendor found>"

                # ---------- Iterate through the list to find a matching prefix
                foreach ($pair in $macVendorList) {
                    if ($normalizedMac.StartsWith($pair.MacPrefix)) {
                        $vendorName = $pair.VendorName
                        break;
                    }
                } 

                $resultsQueue.Add("$ipTemp; $arpResult; $vendorName")
            }
        }
    } catch [System.OperationCanceledException] {
        # Expected exception when the jobQueue is marked as completed and empty
    } 
}

# ----- Create start / stop IP as integer ------------------------------------------------------------

$ipStartInt = Ip2Int $ipStart
$ipEndInt = Ip2Int $ipEnd

if ($ipStartInt -gt $ipEndInt) {
    throw "The start IP must be less than or equal to the end IP"
    exit 1
}

$global:currentValue = $ipStartInt

# ----- Create the worker threads ------------------------------------------------------------

write-host "Creating $maxThreads threads"

$hostsToScan = ($ipEndInt - $ipStartInt + 1)

# Create and start runspaces (threads)
$runspaces = 1..$maxThreads | ForEach-Object {
    $runspace = [powershell]::Create().AddScript($scriptBlock).AddArgument($jobQueue).AddArgument($resultsQueue).AddArgument($macVendorList)
    $runspace.BeginInvoke() 
}

# Start time
$startTime = Get-Date

write-host "Scan from:   $ipStart"
write-host "Scan to:     $ipEnd"
write-host "Total hosts: $hostsToScan"

# ----- Put write and read togeather ------------------------------------------------------------
do {
    $nextValue = $null
    $newResult = $null

    #Check for new result and display
    if ($resultsQueue.TryTake([ref]$newResult)) {
        # $newResult | Convert-Decimal2IP
        $newResult
    }

    if ($jobQueue.IsAddingCompleted) {

        # Wait for all runspaces to finish processing
        $runspaces | ForEach-Object { $_.AsyncWaitHandle.WaitOne() } | Out-Null
		
		# End time
		$endTime = Get-Date

        # Since all jobs are processed, mark the results queue as completed
        $resultsQueue.CompleteAdding()

        # Output the results
        $resultsQueue | ForEach-Object { 
            #Write-Output $_ | Convert-Decimal2IP
            Write-Output $_ 
        } 

        break; # break the do..while
        #<----------
    } else {

        # write next jobs on queue
        1..100 | ForEach-Object {

            if ($jobQueue.Count -ge $maxJobsOnJobQueue) {
                return; # break the ForEach-Object
                #<-----------
            }

            $nextValue = Get-NextJob
            if ($nextValue -ne $null) {
                $jobQueue.Add($nextValue)
            } else {
                $jobQueue.CompleteAdding()
                return; # breaks the ForEach-Object
                #<----------
            }
        }
    }

} while ($true)

# Calculate time taken
$timeTaken = New-TimeSpan -Start $startTime -End $endTime

# Calculate hosts per second
$hostsPerSecond = [math]::Round($hostsToScan / $timeTaken.TotalSeconds, 2)
$totalTimeFormatted = "{0:hh\:mm\:ss\.fff}" -f $timeTaken

# Display results
Write-Host "Hosts to scan:  $hostsToScan"
Write-Host "Start Time:     $($startTime.ToString())"
Write-Host "End Time:       $($endTime.ToString())"
Write-Host "Total Time:     $totalTimeFormatted"
Write-Host "Hosts / second: $hostsPerSecond"

